/** client.c
 *
 * This is meant to be a C client to the bzr service program.
 * Basically, it is used to have a compiled front-end which
 * takes no time to load, and then makes requests against the
 * python service, which stays memory resident.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>

#ifdef __BIG_ENDIAN__
#define HTON_4(x) x
#else
#define HTON_4(x) ((((x) & 0x000000FF) << 24) \
		         & (((x) & 0x0000FF00) <<  8) \
		         & (((x) & 0x00FF0000) >>  8) \
		         & (((x) & 0xFF000000) >> 24))
#endif
#define NTOH_4 HTON_4

/* We need a 4-byte word here.
 * size_t is probably 8-bytes on 64-bit platforms
 * This needs to be made 64-bit compatible.
 */
typedef unsigned int uint32;

int
read_uint(int sock, uint32 *local)
{
	uint32 big_endian = 0;
	ssize_t count = 0;

	count = read(sock, &big_endian, 4);
	if (count != 4) {
		/* Failure */
		return 0;
	}
	*local = NTOH_4(big_endian);
	return 1;
}

void
write_uint(int sock, uint32 local)
{
	uint32 big_endian = 0;

	big_endian = HTON_4(local);
	write(sock, &big_endian, 4);
}

void
write_str(int sock, char const *str)
{
	size_t len = strlen(str);
	write_uint(sock, len);
	write(sock, str, len);
}

char *
read_str(int sock, uint32 *ret_len)
{
	char * str = NULL;
	uint32 len = 0;
	ssize_t count = 0;
	uint32 total = 0;

	if (!read_uint(sock, &len)) {
		/* Failure */
		return 0;
	}
	if (ret_len) {
		*ret_len = len;
	}
	str = (char *)malloc(len+1);
	memset(str, 0, len+1);

	while (total < len) {
		count = read(sock, str+total, len);
		if (count == 0) {
			/* EOF reached before len, Failure */
			free(str);
			return 0;
		}
		total += count;
	}
	return str;
}

int
main(int argc, char **argv)
{
	int sock = 0;
	struct sockaddr_in addr;
	int err = 0;
	char cur_path[8192] = {0};
	uint32 retcode = 0;
	int i = 0;
	char *str = 0;

	addr.sin_family = AF_INET;
	addr.sin_port = 11111;
	addr.sin_addr.s_addr = 0x7F000001; /* 127.0.0.1 */

	getcwd(cur_path, 8192);

	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("Failed to create socket.");
		retcode = 1;
		goto quit;
	}

	err = connect(sock, (struct sockaddr*)&addr, sizeof(addr));
	if (err != 0) {
		perror("Failed to connect to localhost");
		retcode = 2;
		goto quit;
	}

	write_str(sock, cur_path);
	write_uint(sock, argc - 1);
	for (i = 1; i < argc; ++i) {
		write_str(sock, argv[i]);
	}

	if (!read_uint(sock, &retcode)) {
		fprintf(stderr, "We failed to read the return code.");
		retcode = 3;
		goto quit;
	}

	/* The output of the command */
	str = read_str(sock, 0);
	fputs(str, stdout);
	free(str);

  quit:
	close(sock);
	return retcode;
}

